function findPath(%a, %b, %callback, %heuristic, %finder) {
	if (!isObject(%a) || !isObject(%b)) {
		error("ERROR: Invalid targets.");
		return -1;
	}

	if (%heuristic $= "") {
		%heuristic = "euclideanPathCost";
	}
	else if (!isFunction(%heuristic)) {
		error("ERROR: Invalid path cost heuristic.");
		return -1;
	}

	if (%finder $= "") {
		%finder = AStarFinder;
	}
	else if (!isFunction(%finder, "onAdd")) {
		error("ERROR: Invalid path finding method.");
		return -1;
	}

	%a = %a.getID();
	%b = %b.getID();

	%obj = new ScriptObject() {
		class = %finder;

		heuristic = %heuristic;
		callback = %callback;

		a = %a;
		b = %b;
	};

	return %obj;
}

function AStarFinder::onAdd(%this) {
	%this.done = 0;
	%this.result = "";

	if (%this.a == %this.b || %this.a.linkedTo[%this.b]) {
		%this.end(%this.a SPC %this.b);
		return;
	}

	%this.open = %this.a;
	%this.score[%this.a] = call(%this.heuristic, %this.a, %this.b);

	%this.schedule(0, tick);
}

function AStarFinder::tick(%this) {
	cancel(%this.tick);

	if (%this.done) {
		return;
	}

	if (%this.open $= "") {
		%this.end();
		return;
	}

	%count = getWordCount(%this.open);

	for (%i = getWordCount(%this.open) - 1; %i >= 0; %i--) {
		%test = getWord(%this.open, %i);

		if (%best $= "" || %this.score[%test] < %best) {
			%best = %this.score[%test];
			%node = %test;
		}
	}

	if (%node $= "") {
		%this.end();
		return;
	}

	%this.runFor(%node);

	if (!%this.done) {
		%this.tick = %this.schedule(0, tick);
	}
}

function AStarFinder::runFor(%this, %node) {
	if (%node == %this.b) {
		while (%this.from[%node] !$= "") {
			%path = prependItemToList(%path, %node);
			%node = %this.from[%node];
		}

		%this.end(%this.a SPC %path);
	}

	%this.close = appendItemToList(%this.close, %node);
	%this.open = removeItemFromList(%this.open, %node);

	for (%i = 0; %i < %node.linkCount; %i++) {
		%link = %node.link[%i];

		if (!isItemInList(%this.close, %link)) {
			%this.score[%link] = call(%this.heuristic, %link, %this.b);
			%this.from[%link] = %node;

			if (!isItemInList(%this.open, %link)) {
				%this.open = appendItemToList(%this.open, %link);
			}
		}
	}
}

function AStarFinder::end(%this, %result) {
	if (!%this.done) {
		%this.done = 1;
		%this.result = %result;

		if (isFunction(%this.callback)) {
			call(%this.callback, %this, %result);
		}
	}
}

function euclideanPathCost(%a, %b) {
	return vectorDist(%a.position, %b.position);
}

function manhattanPathCost(%a, %b) {
	%x = mAbs(getWord(%a.position, 0) - getWord(%b.position, 0));
	%y = mAbs(getWord(%a.position, 1) - getWord(%b.position, 1));
	%z = mAbs(getWord(%a.position, 2) - getWord(%b.position, 2));

	return %x + %y + %z;
}

function chebyshevPathCost(%a, %b) {
	%x = mAbs(getWord(%a.position, 0) - getWord(%b.position, 0));
	%y = mAbs(getWord(%a.position, 1) - getWord(%b.position, 1));
	%z = mAbs(getWord(%a.position, 2) - getWord(%b.position, 2));

	if (%x > %y && %x > %z) {
		return %z;
	}

	if (%y > %x && %y > %z) {
		return %y;
	}

	return %z;
}

function appendItemToList(%list, %item, %separator) {
	if (%separator $= "") {
		%separator = " ";
	}

	if (%list $= "") {
		return %item;
	}
	else {
		return %list @ %separator @ %item;
	}
}

function prependItemToList(%list, %item, %separator) {
	if (%separator $= "") {
		%separator = " ";
	}

	if (%list $= "") {
		return %item;
	}
	else {
		return %list @ %separator @ %item;
	}
}

function removeItemFromList(%list, %item, %separator) {
	//%list = strReplace(%separator @ %list @ %separator, %separator @ %item @ %separator, %separator);

	//if (%list !$= "" && getSubStr(%list, 0, 1) $= "") {
	//	%list = getSubStr(%list, 1, strLen(%list));
	//}
	//
	//if (%list !$= "" && getSubStr(%list, strLen(%list) - 1, 1) $= "") {
	//	%list = getSubStr(%list, 0, strLen(%list) - 1);
	//}

	return trim(strReplace(%separator @ %list @ %separator, %separator @ %item @ %separator, %separator));
}

function isItemInList(%list, %item, %separator) {
	if (%separator $= "") {
		%separator = " ";
	}

	return strPos(%separator @ %list @ %separator, %separator @ %item @ %separator) >= 0;
}
